
import dtiplayground.dmri.preprocessing as prep
import dtiplayground.dmri.preprocessing.dwi as dwi

import shutil
import yaml,sys,traceback,time
from pathlib import Path

logger=prep.logger.write

def _load_protocol(filename):
    return yaml.safe_load(open(filename,'r'))


def load_configurations(config_dir:str):
    ## reparametrization
    home_dir=Path(config_dir)
    ## Function begins
    config_filename=home_dir.joinpath("config.yml")
    environment_filename=home_dir.joinpath("environment.yml")
    config=yaml.safe_load(open(config_filename,'r'))
    environment=yaml.safe_load(open(environment_filename,'r'))
    return config,environment

def _generate_exec_sequence(pipeline,image_paths:list,output_dir,modules, io_options): ## generate sequence using uuid to avoid the issue from redundant module use
    seq=[]
    after_multi_input=False
    for idx,parr in enumerate(pipeline):
        module_name, options=parr 
        is_multi_input='multi_input' in modules[module_name]['template']['process_attributes']
        is_single_input_allowed="single_input" in modules[module_name]['template']['process_attributes']
        if (is_multi_input and (len(image_paths)<2)):
            if(not is_single_input_allowed):
                raise Exception("Multi_input module needs at least 2 input images. in {}".format(module_name))
        if ((is_multi_input and (len(image_paths)>=2)) or after_multi_input) :
            uid=prep.get_uuid()
            execution={
                "order": idx,
                "id":uid,
                "multi_input":True,
                "module_name":module_name,
                "options": options,
                "image_path": Path(output_dir).joinpath(io_options['output_filename_base']).__str__(),
                "output_base": Path(output_dir).joinpath(io_options['output_filename_base']).__str__(),
               "save": idx+1==len(pipeline) ## is it the final stage? (to save the final output)
            }
            ## save previous results
            if idx>0 and not after_multi_input:
                for idx2,ip in enumerate(image_paths):
                    seq[-idx2-1]['save']=True
            seq.append(execution)
            after_multi_input=True
        else:
            for ip in image_paths:
                uid=prep.get_uuid()
                execution={
                    "order": idx,
                    "id":uid,
                    "multi_input":False,
                    "module_name":module_name,
                    "options": options,
                    "image_path":ip,
                    "output_base": ip,
                    "save": idx+1==len(pipeline)  ## is it the final stage? (to save the final output)
                }
                seq.append(execution)
    return seq 


def _generate_output_directories_mapping(output_dir,exec_sequence): ## map exec sequence uuid to output directory
    # Path(output_dir).mkdir(parents=True,exist_ok=True)
    module_output_dirs={}
    for idx,execution in enumerate(exec_sequence):
        # uid, m,options=parr 
        uid=execution['id']
        m=execution['module_name']
        options=execution['options']
        image_path=execution['image_path']
        order=execution['order']
        module_output_dir=Path(output_dir).joinpath(Path(image_path).stem.split('.')[0]).joinpath("{:02d}_{}".format(order,m))
        module_output_dirs[uid]=str(module_output_dir)
    return module_output_dirs

def default_pipeline_options():
    return {
                 "options":{
                    "overwrite":False, # if result.yml exists and overwrite is false, module skips overall computation except for postProcess
                    # "recompute":False,
                    "write_image":False, # unless module is forced to write (such as BASELINE_Average, or correcting ones)
                    "skip":False
                    }, 
                 "protocol" :{}
            }

class Protocols:
    def __init__(self,config_dir,modules=None,*args,**kwargs):
        self.image_paths=[]
        self.protocol_filename=None
        self.rawdata=None
        self.pipeline=None
        self.io={}
        self.io_options={}
        self.version=None
        self.config_dir=config_dir 

        #Image data 
        self.original_image_information=None
        self.original_image_format='nrrd'
        self.images=[]
        self.image_cache={} # cache for the previous results

        #Execution variables
        self.template_filename=Path(__file__).resolve().parent.joinpath("templates/protocol_template.yml")
        self.modules=modules
        self.previous_process=None #this is to ensure to access previous results (image and so on)
        self.software_info=None # binary path of softwares (such as fsl)
        self.num_threads=4 # number of threads to use 
        self.global_variables={} # global variables to track from each module (arbitrary key-value dict)
        if 'global_vars' in kwargs:
            self.global_variables.update(kwargs['global_vars'])
        #Module related
        self.config,self.environment=load_configurations(self.config_dir)

        #output
        self.result_history={}
        self.output_dir=None
        
        self.setSoftwareInfo()

    def loadImages(self, image_paths,b0_threshold=10):
        self.image_paths=list(map(lambda x:str(Path(x).absolute()),image_paths))
        #print(self.image_paths)
        for ip in self.image_paths:
            logger("Loading original image : {}".format(str(ip)),prep.Color.PROCESS)
            img=dwi.DWI(str(ip))
            self.result_history[ip]=[{"output":{"image_path": str(Path(ip).absolute()),
                                             "image_information": img.information,
                                             "image_object" : id(img)}}]
            img.setB0Threshold(b0_threshold)
            img.getGradients()
            self.images.append(img)
        self.original_image_information = self.images[0].information
        self.original_image_format = self.images[0].image_type

    def setOutputDirectory(self, output_dir=None):
        if output_dir is None:
            self.output_dir=Path(self.getImagePath()).parent
        else:
            self.output_dir=str(Path(output_dir).absolute())
        self.io['output_directory']=str(self.output_dir)

    def setSoftwareInfo(self, paths:object=None):
        softwares=None
        if paths is None:
            spaths=[Path(self.config_dir).joinpath('software_paths.yml'),Path(__file__).resolve().parent.parent.joinpath('common/data/software_paths.yml')]
            for p in spaths:
                if Path(p).exists():
                    softwares=yaml.safe_load(open(p,'r'))
                    self.software_info=softwares 
                    self.num_threads=softwares['parameters']['num_max_threads']
                    break 
        if softwares is None: raise Exception("Software information is required")
        return softwares is not None 

    def getSoftwareInfo(self):
        return self.software_info 

    def setNumThreads(self,nth:int):
        assert(nth>0)
        self.num_threads=nth 
        self.software_info['parameters']['num_max_threads']=nth
        self.io['num_threads']=nth  

    def getProtocols(self):
        proto={
            'version' : self.version,
            'io' : self.io,
            'pipeline': self.pipeline
        }
        return proto

    def writeProtocols(self,filename):
        self.rawdata=self.getProtocols()
        yaml.safe_dump(self.rawdata,open(filename,'w'))


    def loadProtocols(self,filename):
        try:
            self.rawdata=_load_protocol(filename)
            self.version=self.rawdata['version']
            self.pipeline=self.furnishPipeline(self.rawdata['pipeline'])
            self.io=self.rawdata['io']
            if 'no_output_image' not in self.io:
                self.io['no_output_image']= False
            if 'output_format' not in self.io:
                self.io['output_format']=None
            if 'baseline_threshold' not in self.io:
                self.io['baseline_threshold']=10
            self.protocol_filename=filename
            return True
        except Exception as e:
            logger("Exception occurred : {}".format(str(e)))
            traceback.print_exc()
            return False

    def loadModules(self,pipeline:list,user_module_paths:list, **options):
        mod_names=[x for x in pipeline]
        modules=prep.modules._load_modules(user_module_paths=user_module_paths,module_names=mod_names, **options)
        modules=prep.modules.check_module_validity(modules, self.environment, self.config_dir)
        self.modules=modules

        return self.modules 

    def setModules(self,modules):
        self.modules=modules 

    def addPipeline(self,modulename,options={},index=-1):
        opt=default_pipeline_options()
        if 'options' in options:
            opt['options'].update(options['options'])
        default_protocol=getattr(self.modules[modulename]['module'],modulename)().generateDefaultProtocol(self.images[0])
        opt['protocol'].update(default_protocol)
        if 'protocol' in options:
            opt['protocol'].update(options['protocol'])
        self.pipeline.insert(index, [modulename,opt])

    def makeDefaultProtocols(self,pipeline=None,template=None,options={}):
        self.checkImage()
        logger("Default protocols are being generated using image information",prep.Color.PROCESS)
        if template==None:
            template=yaml.safe_load(open(self.template_filename,'r'))

        ### generate default protocols
        self.io={}
        self.version=template['version']
        for k,elm in template['options']['io'].items():
            self.io[k]=elm['default_value']
        if self.output_dir is not None:
            self.io['output_directory']=str(self.output_dir)
        if 'baseline_threshold' in options:
            self.io['baseline_threshold']=options['baseline_threshold']
        self.io['output_format']=None
        if 'output_format' in options:
            self.io['output_format']=options['output_format']
        self.io['no_output_image']= False
        if 'no_output_image' in options:
            self.io['no_output_image']=options['no_output_image']
        if pipeline is not None:
            self.pipeline=self.furnishPipeline(pipeline)
        else:
            self.pipeline=self.furnishPipeline(template['options']['execution']['pipeline']['default_value'])
        logger("Default protocols are generated.",prep.Color.OK)

    def furnishPipeline(self,pipeline):
        self.checkImage()
        module_names=[]
        for idx,parr in enumerate(pipeline):
            if not isinstance(parr, list):
                mod_name = parr
            else:
                mod_name , _ = parr
            module_names.append(mod_name)
        opts={
                "software_info": self.getSoftwareInfo(),
                "global_variables" : self.global_variables
             }
        if 'baseline_threshold' in self.io:
            opts['baseline_threshold'] = self.io['baseline_threshold']
        self.loadModules(module_names,user_module_paths=self.config['user_module_directories'],**opts)
        new_pipeline=[]
        for idx,parr in enumerate(pipeline):
            mod_name = None
            if not isinstance(parr, list):
                mod_name = parr
            else:
                mod_name , _ = parr
            default_options=default_pipeline_options()

            default_protocol=getattr(self.modules[mod_name]['module'],mod_name)(str(self.config_dir), **opts).generateDefaultProtocol(self.images[0])
            default_options['protocol'].update(default_protocol)
            if not isinstance(parr, list):
                new_pipeline.append([parr,default_options])
            else:
                p,opt = parr 
                if "options" in opt:
                    if opt["options"]==None : opt["options"]={}
                    default_options['options'].update(opt['options'])
                if "protocol" in opt:
                    if opt["protocol"]==None : opt["protocol"]={}
                    default_options['protocol'].update(opt['protocol'])
                newopt=default_options
                new_pipeline.append([p,newopt])
        return new_pipeline 


    def checkRunnable(self):
        logger("Checking runability ...",prep.Color.PROCESS)
        self.checkImage()
        self.checkPipeline()
        self.checkDependencies()

    def checkImage(self):
         if len(self.images) ==0: 
            logger("[ERROR] Image is not loaded.",prep.Color.ERROR)
            raise Exception("Image is not set")       
    def checkPipeline(self):
        if self.pipeline is None:
            logger("[ERROR] Protocols are not set.",prep.Color.ERROR)
            raise Exception("Pipe line error")
    def checkDependencies(self):
        for parr in self.pipeline:
            p, options = parr 
            if not self.modules[p]['valid']: 
                msg=self.modules[p]['validity_message']
                logger("[ERROR] Dependency is not met for the module : {} , {}".format(p,msg),prep.Color.WARNING)
                raise Exception("Module {} is not configured correctly.".format(p))    

    def getBaseFilename(self,org_filename):
        fname = Path(org_filename).stem
        potential_kwds=['ap','pa','hf','fh','si','is','lr','rl']
        potential_kwds += list(map(lambda x: x.upper(),potential_kwds))
        for kwd in potential_kwds:
            splitter="_{}".format(kwd)
            if splitter in fname:
                return fname.split(splitter)[0]
        return fname.split('.')[0]

    def loadGlobalVariables(self):
        gv_filename=Path(self.output_dir).joinpath('global_variables.yml')
        if gv_filename.exists():
            return yaml.safe_load(open(gv_filename,'r'))
        else:
            return {}
    def writeGlobalVariables(self):
        gv_filename=Path(self.output_dir).joinpath('global_variables.yml')
        yaml.safe_dump(self.global_variables,open(gv_filename,'w'))

    @prep.measure_time
    def runPipeline(self,options={}):
        try:
            if 'execution_id' in options: logger("Execution ID : {}".format(options['execution_id']))
            self.checkRunnable()
            self.processes_history=[]
            self.io_options['output_filename_base']=self.getBaseFilename(self.images[0].filename)
            if options['output_file_base']:
                self.io_options['output_filename_base']=options['output_file_base']
            execution_sequence = _generate_exec_sequence(self.pipeline,self.image_paths,self.output_dir,self.modules, self.io_options)
            Path(self.output_dir).mkdir(parents=True,exist_ok=True)
            output_dir_map=_generate_output_directories_mapping(self.output_dir,execution_sequence)
            protocol_filename=Path(self.output_dir).joinpath('protocols.yml').__str__()
            logger("Writing protocol file to : {}".format(protocol_filename),prep.Color.PROCESS)
            self.writeProtocols(protocol_filename)
            ## print pipeline
            logger("PIPELINE",prep.Color.INFO)
            logger(yaml.safe_dump(self.io),prep.Color.DEV)
            logger(yaml.safe_dump(self.pipeline),prep.Color.DEV)
            ## run pipeline
            opts={
                    "software_info": self.getSoftwareInfo(),
                    "baseline_threshold" : self.io['baseline_threshold'],
                    "global_variables" : self.global_variables
                 }
            forced_overwrite=False
            self.global_variables.update(self.loadGlobalVariables())
            for idx,execution in enumerate(execution_sequence):
                # uid, p, options=parr 
                uid=execution['id']
                p=execution['module_name']
                options=execution['options']
                image_path=execution['image_path']
                output_base=execution['output_base']
                save=execution['save']

                if image_path not in self.result_history:
                    self.result_history[image_path]=[]
                bt=time.time()
                logger("-----------------------------------------------",prep.Color.BOLD)
                logger("Processing [{0}/{1}] : {2}".format(idx+1,len(execution_sequence),p),prep.Color.BOLD)
                logger("Filename: {}".format(image_path),prep.Color.BOLD)    
                logger("-----------------------------------------------",prep.Color.BOLD)
                Path(output_dir_map[uid]).mkdir(parents=True,exist_ok=True)
                logger("Output directory : {}\n".format(str(output_dir_map[uid])),prep.Color.DEV)
                m=getattr(self.modules[p]['module'], p)(self.config_dir, **opts)
                m.setOptionsAndProtocol(options)

                logger(yaml.safe_dump(m.getTemplate()['process_attributes']),prep.Color.DEV)
                logger(yaml.safe_dump(m.getOptions()),prep.Color.DEV)
                logger(yaml.safe_dump(m.getProtocol()),prep.Color.DEV)
                if m.getOptions()['skip']:
                    forced_overwrite=True 
                    logger("SKIPPING THIS",prep.Color.INFO)
                    continue

                m.initialize(self.result_history,image_path,output_dir=output_dir_map[uid])
                success=False
                resultfile_path=Path(output_dir_map[uid]).joinpath('result.yml')

                ### if result file is exist, just run post process and continue with previous information

                if m.getOptions()['overwrite']:
                    forced_overwrite=True 

                if resultfile_path.exists() and not m.getOptions()['overwrite'] and not forced_overwrite:
                    result_temp=yaml.safe_load(open(resultfile_path,'r'))
                    logger("Result file exists, just post-processing ...",prep.Color.INFO+prep.Color.BOLD)
                    m.postProcess(result_temp,opts)
                    success=True
                else: # in case overwriting or there is no result.yml file
                    outres=m.run(opts,global_vars=self.global_variables)
                    success=outres['success']
                if not success:
                    logger("[ERROR] Process failed in {}".format(p),prep.Color.ERROR) 
                    raise Exception("Process failed in {}".format(p))
                self.global_variables.update(m.getGlobalVariables())
                self.writeGlobalVariables()
                self.previous_process=m  #this is for the image id reference
                self.result_history[image_path] =m.getResultHistory()
                self.image_cache[image_path]=m.image 
                for intermediary_file in m.getOutputFiles():
                    srcfilepath = intermediary_file['source']
                    postfix = intermediary_file['postfix']
                    output_stem = Path(output_base).name.split('.')[0]+"_{}".format(postfix)

                    ext=Path(srcfilepath).suffix
                    if ".nrrd" in srcfilepath.lower():
                        ext=".nrrd"
                    elif ".nii.gz" in srcfilepath.lower():
                        ext=".nii.gz"
                    else:
                        ext=Path(srcfilepath).suffix
                    filename="{}{}".format(output_stem,ext)
                    output_path = Path(self.output_dir).joinpath(filename)
                    logger("Saving intermediary files from {} to {}".format(srcfilepath, output_path),prep.Color.PROCESS)
                    shutil.copy(srcfilepath, output_path)

                et=time.time()-bt
                self.result_history[image_path][-1]['processing_time']=et                   
                logger("[{}] Processed time : {:.2f}s".format(p,et),prep.Color.DEV)
                if save and not self.io['no_output_image']: ### for the last, dump image and informations
                    ## Save final Qced image
                    logger("Preparing final output ... ",prep.Color.PROCESS)
                    stem=Path(output_base).name.split('.')[0]+"_QCed"
                    ext='.nii.gz'
                    if self.io['output_format'] is None: self.io['output_format']=self.original_image_format
                    if self.io['output_format'] =='nrrd' : ext='.nrrd'
                    final_filename=Path(self.output_dir).joinpath(stem).__str__()+ext
                    final_gradients_filename=Path(self.output_dir).joinpath(Path(output_base).stem.split('.')[0]).joinpath('output_gradients.yml').__str__()
                    final_information_filename=Path(self.output_dir).joinpath(Path(output_base).stem.split('.')[0]).joinpath('output_image_information.yml').__str__()
                    
                    if not Path(final_filename).exists() or idx+1==len(execution_sequence):
                        m.image.writeImage(final_filename,dest_type=self.io['output_format'])
                        m.image.dumpGradients(final_gradients_filename)
                        m.image.dumpInformation(final_information_filename)

            logger(yaml.safe_dump(execution_sequence),prep.Color.INFO)
            return self.result_history

        except Exception as e:
            logger("Exception occurred in runPipeline {}".format(str(e)),prep.Color.ERROR)
            tbstr=traceback.format_exc()
            logger(tbstr,prep.Color.ERROR)
            return None
        finally:
            with open(Path(self.output_dir).joinpath('result_history.yml'),'w') as f:
                yaml.safe_dump(self.result_history,f)




from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class ImageMath(ExternalToolWrapper):
    def __init__(self,binary_path= None, **kwargs):
        super().__init__(binary_path, **kwargs)
        self.binary_path=None
        if binary_path is not None:
            self.binary_path=binary_path
        elif 'softwares' in kwargs:
            self.binary_path=kwargs['softwares']['ImageMath']['path']

        self.dev_mode=False

    def rescale(self,inputfile, outputfile , rescale=[0,10000]):
        arguments=[inputfile,'-outfile',outfile,'-rescale',",".join(rescale)]
        self.setArguments(arguments)
        return self.execute(arguments)

    def normalize(self,inputfile,outputfile,referencefile):
        arguments=[inputfile,'-outfile',outputfile,'-matchHistogram',referencefile]
        self.setArguments(arguments)
        return self.execute(arguments)

    def average(self,inputfile,outputfile,files_to_average:list):
        arguments=[inputfile,'-outfile',outputfile,'-avg']+files_to_average
        self.setArguments(arguments)
        return self.execute(arguments)




    



import os
import time
import sys
import json 
import argparse
import csv
import shutil
import threading
import traceback
from copy import deepcopy
from pathlib import Path 
import yaml 

import dtiplayground.dmri.atlasbuilder.utils as utils 
import dtiplayground.dmri.common
import dtiplayground.dmri.common.tools as ext_tools 

logger=dtiplayground.dmri.common.logger.write


## builder class
class AtlasBuilder(object):
    def __init__(self,*args,**kwargs):
        self.configuration=None  ## configuration (build, params, ...)
        self.tools={}  #external tools

    def configure(self,output_dir,config_path,hbuild_path,greedy_params_path,buildsequence_path=None,node=None):

        ### assertions
        assert(Path(config_path).exists())
        assert(Path(hbuild_path).exists())
        assert(Path(greedy_params_path).exists())

        ### init output directories
        projectPath=Path(output_dir).absolute().resolve(strict=False)
        scriptPath=projectPath.joinpath("scripts")
        commonPath=projectPath.joinpath('common')
        configPath=Path(config_path)
        hbuildPath=Path(hbuild_path)
        greedyParamsPath=Path(greedy_params_path)

        ### generate directories
        projectPath.mkdir(parents=True,exist_ok=True)
        commonPath.mkdir(parents=True,exist_ok=True)

        ### copy greedy params
        shutil.copy(greedyParamsPath, commonPath)

        ### generate build sequence
        buildSequence=[]
        hbuild={}
        deformSequence=[]
        numThreads=1
        if buildsequence_path is None:
            hbuild={}
            with open(hbuildPath,'r') as f:
                if hbuildPath.suffix == '.yml':
                    hbuild=yaml.safe_load(f)
                elif hbuildPath.suffix=='.json':
                    hbuild=json.load(f)
                else:
                    raise Exception("No supported file, .json or .yml can be accepted")
                hbuildPath=commonPath.joinpath('h-build.yml')
                yaml.dump(hbuild,open(hbuildPath,'w'))

            config={}
            with open(configPath,'r') as f:
                if configPath.suffix == '.yml':
                    config=yaml.safe_load(f)
                elif configPath.suffix == '.json':
                    config=json.load(f)
                else:
                    raise Exception("No supported file, .json or .yml can be accepted")
            configPath=commonPath.joinpath('config.yml')
            config['m_OutputPath']=str(projectPath)
            yaml.dump(config,open(configPath,'w'))

            hbuild["config"]=config
            hbuild['config']['m_GreedyAtlasParametersTemplatePath']=str(commonPath.joinpath('GreedyAtlasParameters.xml'))

            initSequence=utils.parse_hbuild(hbuild,root_path=projectPath,root_node=node)
            buildSequence=utils.furnish_sequence(hbuild,initSequence)

            #save sequence 
            with open(commonPath.joinpath('build_sequence.yml'),'w') as f:
                yaml.dump(buildSequence,f)
        else:
            with open(buildsequence,'r') as f:
                buildSequence=yaml.safe_load(f)
        
        numThreads=max(int(buildSequence[0]["m_NbThreadsString"]),1)
        logger("Loading external tool settings",dtiplayground.dmri.common.Color.INFO)
        ### init external toolset
        tool_list=['ImageMath','ResampleDTIlogEuclidean','CropDTI','DTIProcess','BRAINSFit','GreedyAtlas','DTIAverage','DTIReg','UNU','ITKTransformTools']
        tool_pairs=list(zip(tool_list,config['m_SoftPath']))
        tool_instances=list(map(lambda x: getattr(ext_tools,x[0])(x[1]),tool_pairs))
        self.tools=dict(zip(tool_list,tool_instances))
        # generate scaffolding directories 
        utils.generate_directories(projectPath,buildSequence)

        with open(commonPath.joinpath('initial_sequence.yml'),'w') as f:
            yaml.dump(initSequence,f,indent=4)

        ## generate deformation field map
        deformInitSequence=utils.generate_deformation_track(initSequence,node=hbuild['project']['target_node'])
        deformSequence=utils.furnish_deformation_track(deformInitSequence,projectPath,buildSequence)
        inverseDeformSequence=utils.invert_deformation_track(deformSequence)

        with open(commonPath.joinpath('deformation_track.yml'),'w') as f:
            yaml.dump(deformSequence,f)
        with open(commonPath.joinpath('deformation_track_inverted.yml'),'w') as f:
            yaml.dump(inverseDeformSequence,f)   

        output={
            "buildSequence" : buildSequence,
            "hbuild":hbuild,
            "config":config,
            "deformInitSequence":deformInitSequence,
            "deformSequence":deformSequence,
            "inverseDeformSequence":inverseDeformSequence,
            "projectPath": projectPath,
            "node":node 
        }
        self.configuration=output
        return output

    @dtiplayground.dmri.common.measure_time
    def build(self):
        assert(self.configuration is not None)
        configuration=self.configuration
        buildSequence=configuration['buildSequence']
        hbuild=configuration['hbuild']
        projectPath=configuration['projectPath']
        config=configuration['config']
        numThreads=max(1,int(config["m_NbThreadsString"]))

        ### atlas build begins (to be multiprocessed)
        logger("\n=============== Main Script ================")
        ## threading
        completedAtlases=[] #entry should be the node name 
        runningAtlases=[] # should have length less or equal than numTheads, entry is the node name
        failedAtlases=[]

        def buildAtlas(conf,rt,ct,ft): # rt : list of running threads, ct : list of completed threads, ft: failed threads
            try:
              prjName=conf["m_NodeName"]
              rt.append(prjName)
              self.preprocess(conf)
              self.build_atlas(conf)  
              rt.remove(prjName)
              ct.append(prjName)
            except Exception as e:
              logger("Exception at {}".format(str(e)))
              msg=traceback.format_exc()
              logger("{}".format(msg))
              rt.remove(prjName)
              ft.append(prjName)
              

        numNodes=len(buildSequence)
        threads=[]
        while len(completedAtlases) < numNodes:
            if len(runningAtlases) < numThreads and len(buildSequence)>0:
                if utils.dependency_satisfied(hbuild,buildSequence[0]["m_NodeName"],completedAtlases):
                    cfg=buildSequence.pop(0)
                    utils.generate_results_csv(cfg)
                    th=threading.Thread(target=buildAtlas,args=(cfg,runningAtlases,completedAtlases,failedAtlases))
                    threads.append(th)
                    th.start()
            if len(failedAtlases)>0:
              logger("There is a failed thread.Exiting...",dtiplayground.dmri.common.Color.ERROR)
              raise Exception("Error occurred in one of the threads")
            time.sleep(1.0)

        ## Postprocess
        self.postprocess()
        return True

    @dtiplayground.dmri.common.measure_time
    def preprocess(self,cfg):    
        config=cfg 
        ext_tools=deepcopy(self.tools) ### for thread safe

        logger("\n============ Pre processing =============")
        
        # Files Paths
        allcases = config['m_CasesPath']
        allcasesIDs = config['m_CasesIDs'] 
        OutputPath= Path(config['m_OutputPath']).joinpath("1_Affine_Registration")
        AtlasScalarMeasurementref=None 
        overwrite= config['m_Overwrite']==1
        needToCrop= config['m_NeedToBeCropped']==1

        AtlasScalarMeasurementref= config['m_TemplatePath'] 
        if config['m_RegType']==1:
          AtlasScalarMeasurementref= OutputPath.joinpath(config['m_CasesIDs'][0] + '_'+ config['m_ScalarMeasurement'] + ".nrrd").__str__() #"/ImageTest1_FA.nrrd" #the reference will be the first case for the first loop
        
        # Create directory for temporary files
        OutputPath.mkdir(parents=True,exist_ok=True)
        logger("\n=> Creation of the affine directory = " + OutputPath.__str__())

        # Creating template by processing Case 1 DTI
        ScalarMeasurement=config['m_ScalarMeasurement']

        if config['m_RegType']==0:
          # Rescaling template
          RescaleTemp= OutputPath.joinpath(config['m_ScalarMeasurement'] + "Template_Rescaled.nrrd").__str__()
          if overwrite or (not utils.CheckFileExists(RescaleTemp,0,"")):
            sp_out=ext_tools['ImageMath'].rescale(AtlasScalarMeasurementref,RescaleTemp,rescale=[0,10000])
          else : logger("=> The file \\'" + RescaleTemp + "\\' already exists so the command will not be executed")
          AtlasScalarMeasurementref= RescaleTemp
        else:
        # Filter case 1 DTI
          FilteredDTI= OutputPath.joinpath(config['m_CasesIDs'][0] +"_filteredDTI.nrrd").__str__()
          if overwrite or (not utils.CheckFileExists(FilteredDTI, 0, "" + config["m_CasesIDs"][0] + "" ) ):
            sp_out=ext_tools['ResampleDTIlogEuclidean'].filter_dti(allcases[0],FilteredDTI,'zero')
          else : logger("=> The file \'" + FilteredDTI + "\' already exists so the command will not be executed")

          # Cropping case 1 DTI
          if needToCrop:
            croppedDTI = OutputPath.joinpath(config['m_CasesIDs'][0] + "_croppedDTI.nrrd").__str__()
            if overwrite or (not utils.CheckFileExists(croppedDTI, 0, "" + config['m_CasesIDs'][0] + "" )):
              sp_out=ext_tools['CropDTI'].crop(FilteredDTI,croppedDTI,size=config['m_CropSize'])
            else: logger("=> The file '" + croppedDTI + "' already exists so the command will not be executed")

          # Generating case 
          DTI= allcases[0]
          if needToCrop:
            DTI= OutputPath.joinpath(config['m_CasesIDs'][0]+"_croppedDTI.nrrd").__str__()

          ScalarMeasurement= OutputPath.joinpath(config['m_CasesIDs'][0] + "_" + config['m_ScalarMeasurement']+".nrrd").__str__()
          if overwrite or (not utils.CheckFileExists(ScalarMeasurement, 0, config["m_CasesIDs"][0] )) :
            sp_out=ext_tools['DTIProcess'].measure_scalars(DTI,ScalarMeasurement,scalar_type=config['m_ScalarMeasurement'])
          else : logger("=> The file \'" + ScalarMeasurement + "\' already exists so the command will not be executed")
            
        # Affine Registration and Normalization Loop
        n = 0
        while n <= config['m_nbLoops'] : 
          # if not os.path.isdir(OutputPath.joinpath("Loop" + str(n)).__str__()):
          OutputPath.joinpath("Loop"+str(n)).mkdir(exist_ok=True)
          # Cases Loop
          case= 0
          if config["m_RegType"]==1: 
            case = (n==0) # (n==0) -> bool: =1(true) =0(false) : the first case is the reference for the first loop so it will not be normalized or registered (it is cropped and FAed before the loop)

          while case < len(allcases):
            if n==0: # Filtering and Cropping DTI and Generating FA are only part of the first loop
               # Filter DTI
              # ResampleDTIlogEuclidean does by default a correction of tensor values by setting the negative values to zero
              FilteredDTI= OutputPath.joinpath(allcasesIDs[case] + "_filteredDTI.nrrd").__str__()
              if overwrite or (not Path(FilteredDTI).exists()):
                sp_out=ext_tools['ResampleDTIlogEuclidean'].filter_dti(allcases[case],FilteredDTI,correction='zero')
              else: logger("=> The file \'" + FilteredDTI + "\' already exists so the command will not be executed")
              if needToCrop:
                croppedDTI=OutputPath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()
                if overwrite or (not Path(croppedDTI).exists()):
                  sp_out=ext_tools['CropDTI'].crop(FilteredDTI,croppedDTI,size=config['m_CropSize'])
                else: logger("=> The file \'" + croppedDTI + "\' already exists so the command will not be executed")
              # Generating FA/MD.
              DTI= allcases[case]
              if needToCrop:
                DTI=OutputPath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()

              ScalarMeasurement= OutputPath.joinpath(allcasesIDs[case] + "_" + config["m_ScalarMeasurement"] + ".nrrd").__str__()
              if overwrite or (not Path(ScalarMeasurement).exists()):
                sp_out=ext_tools['DTIProcess'].measure_scalars(DTI,ScalarMeasurement,scalar_type=config['m_ScalarMeasurement'])
              else: logger("=> The file \'" + ScalarMeasurement + "\' already exists so the command will not be executed")
            # Normalization
            ScalarMeasurement= OutputPath.joinpath(allcasesIDs[case] + "_"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            NormScalarMeasurement= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_Norm"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            if overwrite or (not Path(NormScalarMeasurement).exists()):
              sp_out=ext_tools['ImageMath'].normalize(ScalarMeasurement,NormScalarMeasurement,AtlasScalarMeasurementref)
            else: logger("=> The file \'" + NormScalarMeasurement + "\' already exists so the command will not be executed")

            # Affine registration with BrainsFit
            NormScalarMeasurement= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_Norm"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            LinearTranstfm= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_LinearTrans.txt").__str__()
            LinearTrans= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_LinearTrans_"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            InitLinearTransTxt= OutputPath.joinpath(allcasesIDs[case] + "_InitLinearTrans.txt").__str__()
            InitLinearTransMat= OutputPath.joinpath(allcasesIDs[case] + "_InitLinearTrans.mat").__str__()
            InitTrans=InitLinearTransMat
            if n==0 and utils.CheckFileExists( InitLinearTransMat, case, allcasesIDs[case] ) and utils.CheckFileExists( InitLinearTransTxt, case, allcasesIDs[case] ):
              logger("[WARNING] Both \'" + allcasesIDs[case] + "_InitLinearTrans.mat\' and \'" + allcasesIDs[case] + "_InitLinearTrans.txt\' have been found. The .mat file will be used.")              
            elif n==0 and utils.CheckFileExists( InitLinearTransMat, case, allcasesIDs[case] ) : 
              pass 
            elif n==0 and utils.CheckFileExists( InitLinearTransTxt, case, allcasesIDs[case] ) : 
              InitTrans=InitLinearTransTxt
            else : 
              InitTrans=None
            if overwrite or (not Path(LinearTranstfm).exists()) or (not Path(LinearTrans).exists()):
              sp_out=ext_tools['BRAINSFit'].affine_registration(fixed_path=AtlasScalarMeasurementref,
                                                                 moving_path=NormScalarMeasurement,
                                                                 output_path=LinearTrans ,
                                                                 output_transform_path=LinearTranstfm,
                                                                 initial_transform_path=InitTrans,
                                                                 transform_mode=config['m_BFAffineTfmMode']
                                                                 )
            else: logger("=> The file \'" + LinearTranstfm + "\' already exists so the command will not be executed")

            # Implementing the affine registration
            LinearTranstfm= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_LinearTrans.txt").__str__()
            LinearTransDTI= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_LinearTrans_DTI.nrrd").__str__()
            originalDTI= allcases[case]
            if needToCrop:
              originalDTI= OutputPath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()
            if overwrite or (not Path(LinearTransDTI).exists()):
              sp_out=ext_tools['ResampleDTIlogEuclidean'].implement_affine_registration(originalDTI,LinearTransDTI,LinearTranstfm,AtlasScalarMeasurementref)
            else: logger("=> The file \'" + LinearTransDTI + "\' already exists so the command will not be executed")
            # Generating FA/MA of registered images
            LinearTransDTI= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_LinearTrans_DTI.nrrd").__str__()
            if n == config["m_nbLoops"] : LoopScalarMeasurement= OutputPath.joinpath("Loop"+str(n)).joinpath(allcasesIDs[case] + "_Loop"+ str(n)+"_Final"+config["m_ScalarMeasurement"]+".nrrd").__str__() # the last FA will be the Final output
            else : LoopScalarMeasurement= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            
            if overwrite or (not Path(LoopScalarMeasurement).exists()):
                sp_out=ext_tools['DTIProcess'].measure_scalars(LinearTransDTI,LoopScalarMeasurement,scalar_type=config['m_ScalarMeasurement'])
            else: logger("=> The file \'" + LoopScalarMeasurement + "\' already exists so the command will not be executed")
            case += 1 # indenting cases loop

          # FA/MA Average of registered images with ImageMath
          # if config["m_nbLoops"]!=0:
          if n != int(config["m_nbLoops"]) : # this will not be done for the last lap
            ScalarMeasurementAverage = OutputPath.joinpath("Loop" + str(n)).joinpath("Loop" + str(n) + "_"+config["m_ScalarMeasurement"]+"Average.nrrd").__str__()
            ScalarMeasurementforAVG= OutputPath.joinpath("Loop" + str(n)).joinpath(config["m_CasesIDs"][0] + "_Loop" + str(n) + "_" + config["m_ScalarMeasurement"] + ".nrrd").__str__()
            if config["m_RegType"]==1:
              if n == 0 : ScalarMeasurementforAVG= OutputPath.joinpath(config["m_CasesIDs"][0]+"_"+config["m_ScalarMeasurement"]+".nrrd").__str__()
              else : ScalarMeasurementforAVG= OutputPath.joinpath("Loop" + str(n)).joinpath(config["m_CasesIDs"][0]+"_Loop" + str(n) + "_"+config["m_ScalarMeasurement"]+".nrrd").__str__()
            else:
              ScalarMeasurementforAVG= OutputPath.joinpath("Loop" + str(n)).joinpath(config["m_CasesIDs"][0] + "_Loop" + str(n) + "_" + config["m_ScalarMeasurement"] + ".nrrd").__str__()
          
            case = 1
            ScalarMeasurementList=[]
            ScalarMeasurementList.append(ScalarMeasurementforAVG)
            while case < len(allcases):
              ScalarMeasurementforAVG= OutputPath.joinpath("Loop" + str(n)).joinpath(allcasesIDs[case] + "_Loop" + str(n) + "_"+config["m_ScalarMeasurement"]+".nrrd").__str__()                
              ScalarMeasurementList.append(ScalarMeasurementforAVG)
              case += 1             
            if overwrite or (not utils.CheckFileExists(ScalarMeasurementAverage, 0, "")):
              sp_out=ext_tools['ImageMath'].average(ScalarMeasurementList[0],ScalarMeasurementAverage,ScalarMeasurementList[1:])
              AtlasScalarMeasurementref = ScalarMeasurementAverage # the average becomes the reference
            else:
              logger("=> The file '" + ScalarMeasurementAverage + "' already exists so the command will not be executed")
              AtlasScalarMeasurementref = ScalarMeasurementAverage # the average becomes the reference
          n += 1 # indenting main loop

        logger("\n============ End of Pre processing =============")

    @dtiplayground.dmri.common.measure_time
    def build_atlas(self,cfg):
        ext_tools=deepcopy(self.tools) ### for thread safe
        config=cfg

        m_OutputPath=config["m_OutputPath"]
        m_ScalarMeasurement=config["m_ScalarMeasurement"]
        # m_GridAtlasCommand=config["m_GridAtlasCommand"]
        #m_RegType=config["m_RegType"]
        m_Overwrite=config["m_Overwrite"]
        #m_useGridProcess=config["m_useGridProcess"]
        m_SoftPath=config["m_SoftPath"]
        m_nbLoops=config["m_nbLoops"]
        m_TensTfm=config["m_TensTfm"]
        #m_TemplatePath=config["m_TemplatePath"]
        #m_BFAffineTfmMode=config["m_BFAffineTfmMode"]
        m_CasesIDs=config["m_CasesIDs"]
        m_CasesPath=config["m_CasesPath"]
        #m_CropSize=config["m_CropSize"]
        #m_DTIRegExtraPath=config["m_DTIRegExtraPath"]
        m_DTIRegOptions=config["m_DTIRegOptions"]
        # m_GridAtlasCommand=config["m_GridAtlasCommand"]
        # m_GridGeneralCommand=config["m_GridGeneralCommand"]
        m_InterpolLogOption=config["m_InterpolLogOption"]
        m_InterpolOption=config["m_InterpolOption"]
        m_InterpolType=config["m_InterpolType"]
        #m_NbThreadsString=config["m_NbThreadsString"]
        m_NeedToBeCropped=config["m_NeedToBeCropped"]
        m_PythonPath=config["m_PythonPath"]
        m_TensInterpol=config["m_TensInterpol"]
        m_nbLoopsDTIReg=config["m_nbLoopsDTIReg"]

        ### To be removed
        if m_nbLoopsDTIReg is None: m_nbLoopsDTIReg=1

        logger("\n============ Atlas Building =============")

        # Files Paths
        AffinePath=Path(m_OutputPath).joinpath("1_Affine_Registration")
        DeformPath=Path(m_OutputPath).joinpath("2_NonLinear_Registration")
        FinalPath= Path(m_OutputPath).joinpath("3_Diffeomorphic_Atlas")
        FinalResampPath= Path(m_OutputPath).joinpath("4_Final_Resampling")
        FinalAtlasPath= Path(m_OutputPath).joinpath("5_Final_Atlas")

        DeformPath.mkdir(exist_ok=True)
        FinalPath.mkdir(exist_ok=True)
        FinalResampPath.mkdir(exist_ok=True)
        FinalResampPath.joinpath("First_Resampling").mkdir(exist_ok=True)
        FinalResampPath.joinpath("Second_Resampling").mkdir(exist_ok=True)
        FinalResampPath.joinpath("FinalTensors").mkdir(exist_ok=True)
        FinalResampPath.joinpath("FinalDeformationFields").mkdir(exist_ok=True)
        FinalAtlasPath.mkdir(exist_ok=True)
        

# 1 Get Affine information from AffinePath
        # Cases variables
        alltfms=[]
        for i,c in enumerate(m_CasesPath):
          alltfms.append(AffinePath.joinpath("Loop"+str(m_nbLoops)).joinpath(m_CasesIDs[i] + "_Loop" + str(m_nbLoops) +"_LinearTrans.txt").__str__())
        allcases=[]
        if m_NeedToBeCropped==1:
          for i,c in enumerate(m_CasesPath):
            allcases.append(AffinePath.joinpath(m_CasesIDs[i] + "_croppedDTI.nrrd").__str__())
        else:
          for i,c in enumerate(m_CasesPath):
            allcases.append(m_CasesPath[i])
        allcasesIDs=[]
        for i,c in enumerate(m_CasesIDs):
          allcasesIDs.append(m_CasesIDs[i])

# 2 NonLinear_Registration (DeformPath)
        # GreedyAtlas Command
        utils.generateGreedyAtlasParametersFile(config)
        XMLFile= DeformPath.joinpath("GreedyAtlasParameters.xml").__str__()
        ParsedFile= DeformPath.joinpath("ParsedXML.xml").__str__()
        if m_Overwrite==1 or (not utils.CheckFileExists(DeformPath.joinpath("MeanImage.mhd").__str__(), 0, "")):
          sp_out=ext_tools['GreedyAtlas'].compute_deformation_fields(XMLFile,ParsedFile)
          case = 0
          while case < len(allcases): # Renaming
            originalImage=DeformPath.joinpath(allcasesIDs[case] + "_Loop"+str(m_nbLoops)+"_Final"+m_ScalarMeasurement+"DefToMean.mhd").__str__()
            originalHField=DeformPath.joinpath(allcasesIDs[case] + "_Loop"+str(m_nbLoops)+"_Final"+m_ScalarMeasurement+"DefFieldImToMean.mhd").__str__()
            originalInvHField=DeformPath.joinpath(allcasesIDs[case] + "_Loop"+str(m_nbLoops)+"_Final"+m_ScalarMeasurement+"DefFieldMeanToIm.mhd").__str__()
            NewImage= DeformPath.joinpath(allcasesIDs[case] + "_NonLinearTrans_FA.mhd").__str__()
            NewHField=DeformPath.joinpath(allcasesIDs[case] + "_HField.mhd").__str__()
            NewInvHField=DeformPath.joinpath(allcasesIDs[case] + "_InverseHField.mhd").__str__()
            logger("[" + allcasesIDs[case] + "] => Renaming \'" + originalImage + "\' to \'" + NewImage + "\'")
            os.rename(originalImage,NewImage)
            logger("[" + allcasesIDs[case] + "] => Renaming \'" + originalHField + "\' to \'" + NewHField + "\'")
            os.rename(originalHField,NewHField)
            logger("[" + allcasesIDs[case] + "] => Renaming \'" + originalInvHField + "\' to \'" + NewInvHField + "\'")
            os.rename(originalInvHField,NewInvHField)
            case += 1
        else:
          logger("=> The file '" + str(DeformPath.joinpath('MeanImage.mhd')) + " already exists so the command will not be executed")
          # Renaming possible existing old named files from GreedyAtlas\n";
          case = 0
          while case < len(allcases): # Updating old names if needed\n";
            NewImage= DeformPath.joinpath(allcasesIDs[case] + "_NonLinearTrans_" + m_ScalarMeasurement + ".mhd").__str__()
            utils.CheckFileExists(NewImage, case, allcasesIDs[case])
            NewHField=DeformPath.joinpath(allcasesIDs[case] + "_HField.mhd").__str__()
            utils.CheckFileExists(NewHField, case, allcasesIDs[case])
            NewInvHField=DeformPath.joinpath(allcasesIDs[case] + "_InverseHField.mhd").__str__()
            utils.CheckFileExists(NewInvHField, case, allcasesIDs[case])
            case += 1

# 3 Diffeomorphic Atlas (FinalPath)
        # Apply deformation fields 
        case = 0
        while case < len(allcases):
          FinalDTI= FinalPath.joinpath(allcasesIDs[case] + "_DiffeomorphicDTI.nrrd").__str__()
          if m_NeedToBeCropped==1:
            originalDTI= AffinePath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()
          else:
            originalDTI= allcases[case]
          if m_nbLoops==0:
            Ref = AffinePath.joinpath("Loop0").joinpath("Loop0_"+m_ScalarMeasurement+"Average.nrrd").__str__()
          else:
            Ref = AffinePath.joinpath("Loop" + str(m_nbLoops-1)).joinpath("Loop" + str(m_nbLoops-1) + "_" + m_ScalarMeasurement + "Average.nrrd").__str__()

          HField= DeformPath.joinpath(allcasesIDs[case] + "_HField.mhd").__str__()

          if m_Overwrite==1 or (not utils.CheckFileExists(FinalDTI, case, allcasesIDs[case])):
            DiffeomorphicCaseScalarMeasurement = FinalPath.joinpath(allcasesIDs[case] + "_Diffeomorphic"+m_ScalarMeasurement+".nrrd").__str__()
            
            sp_out=ext_tools['ResampleDTIlogEuclidean'].resample(reference_file=Ref,
                                                                   deform_field_file=HField,
                                                                   transformation_file=alltfms[case],
                                                                   moving_file=originalDTI,
                                                                   output_file=FinalDTI,
                                                                   interpolation_type=m_InterpolType,
                                                                   interpolation_option=m_InterpolOption,
                                                                   tensor_interpolation_type=m_TensInterpol,
                                                                   tensor_interpolation_option=m_InterpolLogOption,
                                                                   tensor_transform=m_TensTfm
                                                                   )

            sp_out=ext_tools['DTIProcess'].measure_scalars(inputfile=FinalDTI,
                                                            outputfile=DiffeomorphicCaseScalarMeasurement,
                                                            scalar_type=m_ScalarMeasurement,
                                                            options=['--scalar_float'])

            out_file=FinalPath.joinpath(allcasesIDs[case] + "_DiffeomorphicDTI_float.nrrd").__str__()
            sp_out=ext_tools['UNU'].convert_to_float(FinalDTI,out_file)

          else : logger("=> The file \'" + FinalDTI + "\' already exists so the command will not be executed")
          case += 1

        # DTIaverage computing
        DTIAverage = FinalPath.joinpath("DiffeomorphicAtlasDTI.nrrd").__str__()
        ListForAverage=[]
        case = 0
        while case < len(allcases):
          temp=FinalPath.joinpath(allcasesIDs[case] + "_DiffeomorphicDTI.nrrd").__str__()
          ListForAverage.append(temp)
          case += 1
        sp_out=ext_tools['DTIAverage'].average(ListForAverage,DTIAverage)

        if m_Overwrite==1 or not utils.CheckFileExists(DTIAverage, 0, "") : 
        # Computing some images from the final DTI with dtiprocess
          FA= FinalPath.joinpath("DiffeomorphicAtlasFA.nrrd").__str__()
          cFA= FinalPath.joinpath("DiffeomorphicAtlasColorFA.nrrd").__str__()
          RD= FinalPath.joinpath("DiffeomorphicAtlasRD.nrrd").__str__()
          MD= FinalPath.joinpath("DiffeomorphicAtlasMD.nrrd").__str__()
          AD= FinalPath.joinpath("DiffeomorphicAtlasAD.nrrd").__str__()
          

          sp_out=ext_tools['DTIAverage'].average(ListForAverage,DTIAverage) 
          sp_out=ext_tools['DTIProcess'].measure_scalars(inputfile=DTIAverage,
                                                outputfile=FA,
                                                scalar_type='FA',
                                                options=['--scalar_float','-m',MD,'--color_fa_output',cFA,'--RD_output',RD,'--lambda1_output',AD])
          out_file=FinalPath.joinpath("DiffeomorphicAtlasDTI_float.nrrd").__str__()
          sp_out=ext_tools['UNU'].convert_to_float(DTIAverage,out_file)

        else: logger("=> The file '" + DTIAverage + "' already exists so the command will not be executed")

# 4-1 First_Resampling (FinalResampPath) - to be multi-threaded
        # Computing global deformation fields
        case = 0
        while case < len(allcases):
          if m_NeedToBeCropped==1:
            origDTI= AffinePath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()
          else:
            origDTI= allcases[case]
          GlobalDefField = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_GlobalDisplacementField.nrrd").__str__()
          InverseGlobalDefField = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_GlobalDisplacementField_Inverse.nrrd").__str__()
          FinalDef = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_DeformedDTI.nrrd").__str__()

          BRAINSExecDir = os.path.dirname(m_SoftPath[4])
          dtiprocessExecDir = os.path.dirname(m_SoftPath[3])
          ResampExecDir = os.path.dirname(m_SoftPath[1])
          PathList=[BRAINSExecDir,dtiprocessExecDir,ResampExecDir]

          BRAINSTempTfm = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_" + m_ScalarMeasurement + "_AffReg.txt").__str__()
          ANTSTempFileBase = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_" + m_ScalarMeasurement + "_").__str__()

          if m_Overwrite==1 or not utils.CheckFileExists(FinalDef, case, allcasesIDs[case]) :   
            sp_out=ext_tools['DTIReg'].compute_global_deformation_fields(
                                                                          fixed_volume=DTIAverage,
                                                                          moving_volume=origDTI,
                                                                          scalar_measurement=m_ScalarMeasurement,
                                                                          output_displacement_field=GlobalDefField,
                                                                          output_inverse_displacementField=InverseGlobalDefField,
                                                                          output_volume=FinalDef,
                                                                          initial_affine=alltfms[case],
                                                                          brains_transform=BRAINSTempTfm,
                                                                          ants_outbase=ANTSTempFileBase,
                                                                          program_paths=PathList,
                                                                          dti_reg_options=m_DTIRegOptions,
                                                                          options=[])
            out_file=FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_DeformedDTI_float.nrrd").__str__()
            sp_out=ext_tools['UNU'].convert_to_float(FinalDef,out_file)

          else: logger("=> The file '" + FinalDef + "' already exists so the command will not be executed")
          case += 1

# 4-2 Second_Resampling 

        ### looping begins
        cnt=0
        if m_Overwrite==0:
          for i in range(m_nbLoopsDTIReg):
            if os.path.isdir(FinalResampPath.joinpath("Second_Resampling").joinpath("Loop_"+str(i)).__str__()):
              cnt=i 
          cnt=max(cnt,0)

        while cnt < m_nbLoopsDTIReg:
          logger("-----------------------------------------------------------")
          logger("Iterative Registration cycle %d / %d" % (cnt+1,m_nbLoopsDTIReg) )
          logger("------------------------------------------------------------")
          FinalResampPath.joinpath("Second_Resampling").joinpath("Loop_"+str(cnt)).mkdir(exist_ok=True)
          # dtiaverage recomputing
          IterDir="Loop_"+str(cnt)
          PrevIterDir="Loop_"+str(cnt-1)
          DTIAverage2 = FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasDTI_averaged.nrrd").__str__()

          ListForAverage=[]
          if cnt==0:
            case = 0
            while case < len(allcases):
              temp=FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_DeformedDTI.nrrd").__str__()
              ListForAverage.append(temp)
              case += 1
          else: ### when iterative registration is activated
            case=0
            while case < len(allcases):
              temp=FinalResampPath.joinpath("Second_Resampling").joinpath(PrevIterDir).joinpath(allcasesIDs[case] + "_FinalDeformedDTI.nrrd").__str__()
              ListForAverage.append(temp)
              case += 1 

          if m_Overwrite==1 or not utils.CheckFileExists(DTIAverage2, 0, ""): 
          # Computing some images from the final DTI with dtiprocess
            FA2= FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasFA.nrrd").__str__()
            cFA2= FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasColorFA.nrrd").__str__()
            RD2= FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasRD.nrrd").__str__()
            MD2= FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasMD.nrrd").__str__()
            AD2= FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasAD.nrrd").__str__()

            sp_out=ext_tools['DTIAverage'].average(ListForAverage,DTIAverage2) 
            sp_out=ext_tools['DTIProcess'].measure_scalars(inputfile=DTIAverage2,
                                                  outputfile=FA2,
                                                  scalar_type='FA',
                                                  options=['--scalar_float','-m',MD2,'--color_fa_output',cFA2,'--RD_output',RD2,'--lambda1_output',AD2])

            out_file=FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath("FinalAtlasDTI_float.nrrd").__str__()
            sp_out=ext_tools['UNU'].convert_to_float(DTIAverage2,out_file)    

          else: logger("=> The file '" + DTIAverage2 + "' already exists so the command will not be executed")

          # Recomputing global deformation fields - to be multi-threaded
          SecondResampRecomputed = [0] * len(allcases) # array of 1s and 0s to know what has been recomputed to know what to copy to final folders
          case = 0
          while case < len(allcases):
            if m_NeedToBeCropped==1:
              origDTI2= AffinePath.joinpath(allcasesIDs[case] + "_croppedDTI.nrrd").__str__()
            else:
              origDTI2= allcases[case]
            GlobalDefField2 = FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath(allcasesIDs[case] + "_GlobalDisplacementField.nrrd").__str__()
            InverseGlobalDefField2 = FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath(allcasesIDs[case] + "_InverseGlobalDisplacementField.nrrd").__str__()
            FinalDef2 = FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath(allcasesIDs[case] + "_FinalDeformedDTI.nrrd").__str__()

            BRAINSExecDir = os.path.dirname(m_SoftPath[4])
            dtiprocessExecDir = os.path.dirname(m_SoftPath[3])
            ResampExecDir = os.path.dirname(m_SoftPath[1])
            PathList=[BRAINSExecDir,dtiprocessExecDir,ResampExecDir]

            BRAINSTempTfm = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_" + m_ScalarMeasurement + "_AffReg.txt").__str__()
            ANTSTempFileBase = FinalResampPath.joinpath("First_Resampling").joinpath(allcasesIDs[case] + "_" + m_ScalarMeasurement + "_").__str__()

            if m_Overwrite==1 or not utils.CheckFileExists(FinalDef2, case, allcasesIDs[case])  :
              SecondResampRecomputed[case] = 1
              DTIRegCaseScalarMeasurement = FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath(allcasesIDs[case] + "_FinalDeformed"+m_ScalarMeasurement+".nrrd").__str__()

              sp_out=ext_tools['DTIReg'].compute_global_deformation_fields(
                                                              fixed_volume=DTIAverage2,
                                                              moving_volume=origDTI2,
                                                              scalar_measurement=m_ScalarMeasurement,
                                                              output_displacement_field=GlobalDefField2,
                                                              output_inverse_displacementField=InverseGlobalDefField2,
                                                              output_volume=FinalDef2,
                                                              initial_affine=alltfms[case],
                                                              brains_transform=BRAINSTempTfm,
                                                              ants_outbase=ANTSTempFileBase,
                                                              program_paths=PathList,
                                                              dti_reg_options=m_DTIRegOptions,
                                                              options=[])

              out_file=FinalResampPath.joinpath("Second_Resampling").joinpath(IterDir).joinpath(allcasesIDs[case] + "_FinalDeformedDTI_float.nrrd").__str__()
              sp_out=ext_tools['UNU'].convert_to_float(FinalDef2,out_file)
              sp_out=ext_tools['DTIProcess'].measure_scalars(inputfile=FinalDef2,
                                                outputfile=DTIRegCaseScalarMeasurement,
                                                scalar_type=m_ScalarMeasurement,
                                                options=['--scalar_float'])

            else: logger("=> The file '" + FinalDef2 + "' already exists so the command will not be executed")
            case += 1

          ### Cleanup - delete PrevIterDir
          if cnt > 1:
            PrevPrevIterDir="Loop_"+str(cnt-2)
            DirToRemove = FinalResampPath.joinpath("Second_Resampling").joinpath(PrevPrevIterDir).__str__()
            if os.path.exists(DirToRemove):
              shutil.rmtree(DirToRemove)

          cnt+=1
        # End while cnt < m_nbLoopDTIReg

# 5 FinalAtlasPath
        # Moving final images to final folders
        logger("\n=> Moving final images to final folders")
        case = 0
        LastIterDir="Loop_"+str(m_nbLoopsDTIReg-1)

        while case < len(allcases):
          if SecondResampRecomputed[case] :
            GlobalDefField2 = FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir).joinpath(allcasesIDs[case] + "_GlobalDisplacementField.nrrd").__str__()
            NewGlobalDefField2 = FinalResampPath.joinpath("FinalDeformationFields").joinpath(allcasesIDs[case] + "_GlobalDisplacementField.nrrd").__str__()
            if utils.CheckFileExists(GlobalDefField2, case, allcasesIDs[case]) :
              shutil.copy(GlobalDefField2, NewGlobalDefField2)
            InverseGlobalDefField2 = FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir).joinpath(allcasesIDs[case] + "_InverseGlobalDisplacementField.nrrd").__str__()
            NewInverseGlobalDefField2 = FinalResampPath.joinpath("FinalDeformationFields").joinpath(allcasesIDs[case] + "_InverseGlobalDisplacementField.nrrd").__str__()
            if utils.CheckFileExists(InverseGlobalDefField2, case, allcasesIDs[case]) :
              shutil.copy(InverseGlobalDefField2, NewInverseGlobalDefField2)
            FinalDef2 = FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir).joinpath(allcasesIDs[case] + "_FinalDeformedDTI.nrrd").__str__()
            NewFinalDef2 = FinalResampPath.joinpath("FinalTensors").joinpath(allcasesIDs[case] + "_FinalDeformedDTI.nrrd").__str__()
            if utils.CheckFileExists(FinalDef2, case, allcasesIDs[case]) :
              shutil.copy(FinalDef2, NewFinalDef2)
            FinalDef2f = FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir).joinpath(allcasesIDs[case] + "_FinalDeformedDTI_float.nrrd").__str__()
            NewFinalDef2f = FinalResampPath.joinpath("FinalTensors").joinpath(allcasesIDs[case] + "_FinalDeformedDTI_float.nrrd").__str__()
            if utils.CheckFileExists(FinalDef2f, case, allcasesIDs[case]) :
              shutil.copy(FinalDef2f, NewFinalDef2f)
            DTIRegCaseScalarMeasurement = FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir).joinpath(allcasesIDs[case] + "_FinalDeformed"+m_ScalarMeasurement+".nrrd").__str__()
            NewDTIRegCaseScalarMeasurement = FinalResampPath.joinpath("FinalTensors").joinpath(allcasesIDs[case] + "_FinalDeformed"+m_ScalarMeasurement+".nrrd").__str__()
            if utils.CheckFileExists(DTIRegCaseScalarMeasurement, case, allcasesIDs[case]) :
              shutil.copy(DTIRegCaseScalarMeasurement, NewDTIRegCaseScalarMeasurement)
          case += 1

        # Copy final atlas components to FinalAtlasPath directory

        logger("Copying Final atlas components to " + str(FinalAtlasPath))
        shutil.rmtree(FinalAtlasPath)
        shutil.copytree(FinalResampPath.joinpath("Second_Resampling").joinpath(LastIterDir),FinalAtlasPath)
        shutil.copytree(FinalResampPath.joinpath("FinalDeformationFields"),FinalAtlasPath.joinpath("FinalDeformationFields"))
        shutil.copytree(FinalResampPath.joinpath("FinalTensors"),FinalAtlasPath.joinpath("FinalTensors"))

        logger("\n============ End of Atlas Building =============")   ### after preprocessing, build atlas with non linear registrations

    @dtiplayground.dmri.common.measure_time
    def postprocess(self):
        assert(self.configuration is not None)
        configuration=self.configuration

        deformSequence=configuration['deformSequence']
        inverseDeformSequence=configuration['inverseDeformSequence']
        projectPath=configuration['projectPath']
        config=configuration['config']
        hbuild=configuration['hbuild']
        node=configuration['node']

        ### copy final atals to 'final_atlas' directory
        if node is None:
            src=projectPath.joinpath("atlases").joinpath(hbuild['project']['target_node'])
        else:
            src=projectPath.joinpath("atlases").joinpath(node)
        dst=projectPath.joinpath("final_atlas")
        logger("Copying filed from %s to %s" %(src,dst))
        shutil.rmtree(dst)
        shutil.copytree(src,dst)

        logger("Final atlas copied into %s "% dst)
        ### Concatenate the displacement fields
        logger("\nConcatenating deformation fields")
        utils.ITKTransformTools_Concatenate(config,deformSequence)
        utils.ITKTransformTools_Concatenate_Inverse(config,inverseDeformSequence)
        utils.generate_results_csv_from_deformation_track(deformSequence,projectPath)



